# Generated by Django 4.2.16 on 2024-10-10 07:53

from django.db import migrations, models

import compute_horde.executor_class


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="JobFinishedReceipt",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("job_uuid", models.UUIDField()),
                ("validator_hotkey", models.CharField(max_length=256)),
                ("miner_hotkey", models.CharField(max_length=256)),
                ("validator_signature", models.CharField(max_length=256)),
                (
                    "miner_signature",
                    models.CharField(blank=True, max_length=256, null=True),
                ),
                ("time_started", models.DateTimeField()),
                ("time_took_us", models.BigIntegerField()),
                ("score_str", models.CharField(max_length=256)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobStartedReceipt",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("job_uuid", models.UUIDField()),
                ("validator_hotkey", models.CharField(max_length=256)),
                ("miner_hotkey", models.CharField(max_length=256)),
                ("validator_signature", models.CharField(max_length=256)),
                (
                    "miner_signature",
                    models.CharField(blank=True, max_length=256, null=True),
                ),
                (
                    "executor_class",
                    models.CharField(
                        default=compute_horde.executor_class.ExecutorClass[
                            "spin_up_4min__gpu_24gb"
                        ],
                        max_length=255,
                    ),
                ),
                ("time_accepted", models.DateTimeField()),
                ("max_timeout", models.IntegerField()),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddConstraint(
            model_name="jobstartedreceipt",
            constraint=models.UniqueConstraint(
                fields=("job_uuid",),
                name="receipts_unique_jobstartedreceipt_job_uuid",
            ),
        ),
        migrations.AddConstraint(
            model_name="jobfinishedreceipt",
            constraint=models.UniqueConstraint(
                fields=("job_uuid",),
                name="receipts_unique_jobfinishedreceipt_job_uuid",
            ),
        ),
    ]



import csv
import io

import pytest

from compute_horde.mv_protocol.validator_requests import (
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
)
from compute_horde.receipts.schemas import (
    Receipt,
    ReceiptType,
)
from compute_horde.receipts.transfer import ReceiptFetchError, get_miner_receipts


def receipts_helper(mocked_responses, receipts: list[Receipt], miner_keypair):
    payload_fields = set()
    for payload_cls in [JobStartedReceiptPayload, JobFinishedReceiptPayload]:
        payload_fields |= set(payload_cls.model_fields.keys())

    buf = io.StringIO()
    csv_writer = csv.DictWriter(
        buf,
        [
            "type",
            "validator_signature",
            "miner_signature",
            *payload_fields,
        ],
    )
    csv_writer.writeheader()
    for receipt in receipts:
        match receipt.payload:
            case JobStartedReceiptPayload():
                receipt_type = ReceiptType.JobStartedReceipt
            case JobFinishedReceiptPayload():
                receipt_type = ReceiptType.JobFinishedReceipt
        row = (
            dict(
                type=receipt_type.value,
                validator_signature=receipt.validator_signature,
                miner_signature=receipt.miner_signature,
            )
            | receipt.payload.model_dump()
        )
        csv_writer.writerow(row)

    mocked_responses.get("http://127.0.0.1:8000/receipts/receipts.csv", body=buf.getvalue())
    return get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8000)


def receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair):
    got_receipts = receipts_helper(mocked_responses, receipts, miner_keypair)
    # only the valid receipt should be stored
    assert len(got_receipts) == 1
    assert got_receipts[0] == receipts[0]


def test__get_miner_receipts__happy_path(mocked_responses, receipts, miner_keypair):
    got_receipts = receipts_helper(mocked_responses, receipts, miner_keypair)
    assert len(got_receipts) == 2
    for receipt in receipts:
        got_receipt = [
            x
            for x in got_receipts
            if x.payload.job_uuid == receipt.payload.job_uuid
            and x.payload.__class__ is receipt.payload.__class__
        ][0]
        assert got_receipt == receipt


@pytest.mark.filterwarnings("ignore:Pydantic serializer warnings")
def test__get_miner_receipts__invalid_receipt_skipped(mocked_responses, receipts, miner_keypair):
    """
    Invalidate one receipt payload fields to make it invalid
    """

    receipts[1].payload.miner_hotkey = 0
    receipts[1].payload.validator_hotkey = None
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__miner_hotkey_mismatch_skipped(
    mocked_responses, receipts, miner_keypair, keypair
):
    receipts[1].payload.miner_hotkey = keypair.ss58_address
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__invalid_miner_signature_skipped(
    mocked_responses, receipts, miner_keypair
):
    receipts[1].miner_signature = f"0x{miner_keypair.sign('bla').hex()}"
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__invalid_validator_signature_skipped(
    mocked_responses, receipts, miner_keypair
):
    receipts[1].validator_signature = f"0x{miner_keypair.sign('bla').hex()}"
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__no_receipts(mocked_responses, miner_keypair):
    mocked_responses.get("http://127.0.0.1:8000/receipts/receipts.csv", status=404)
    with pytest.raises(ReceiptFetchError):
        get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8000)

    with pytest.raises(ReceiptFetchError):
        get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8001)



from datetime import timedelta

from django.db import models

from compute_horde.executor_class import DEFAULT_EXECUTOR_CLASS, ExecutorClass
from compute_horde.mv_protocol.validator_requests import (
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
)
from compute_horde.receipts.schemas import Receipt


class ReceiptNotSigned(Exception):
    pass


class AbstractReceipt(models.Model):
    job_uuid = models.UUIDField()
    validator_hotkey = models.CharField(max_length=256)
    miner_hotkey = models.CharField(max_length=256)
    validator_signature = models.CharField(max_length=256)
    miner_signature = models.CharField(max_length=256, null=True, blank=True)

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(fields=["job_uuid"], name="receipts_unique_%(class)s_job_uuid"),
        ]

    def __str__(self):
        return f"job_uuid: {self.job_uuid}"


class JobStartedReceipt(AbstractReceipt):
    executor_class = models.CharField(max_length=255, default=DEFAULT_EXECUTOR_CLASS)
    time_accepted = models.DateTimeField()
    max_timeout = models.IntegerField()

    # https://github.com/typeddjango/django-stubs/issues/1684#issuecomment-1706446344
    objects: models.Manager["JobStartedReceipt"]

    def to_receipt(self) -> Receipt:
        if self.miner_signature is None:
            raise ReceiptNotSigned("Miner signature is required")

        return Receipt(
            payload=JobStartedReceiptPayload(
                job_uuid=str(self.job_uuid),
                miner_hotkey=self.miner_hotkey,
                validator_hotkey=self.validator_hotkey,
                executor_class=ExecutorClass(self.executor_class),
                time_accepted=self.time_accepted,
                max_timeout=self.max_timeout,
            ),
            validator_signature=self.validator_signature,
            miner_signature=self.miner_signature,
        )


class JobFinishedReceipt(AbstractReceipt):
    time_started = models.DateTimeField()
    time_took_us = models.BigIntegerField()
    score_str = models.CharField(max_length=256)

    # https://github.com/typeddjango/django-stubs/issues/1684#issuecomment-1706446344
    objects: models.Manager["JobFinishedReceipt"]

    def time_took(self):
        return timedelta(microseconds=self.time_took_us)

    def score(self):
        return float(self.score_str)

    def to_receipt(self) -> Receipt:
        if self.miner_signature is None:
            raise ReceiptNotSigned("Miner signature is required")

        return Receipt(
            payload=JobFinishedReceiptPayload(
                job_uuid=str(self.job_uuid),
                miner_hotkey=self.miner_hotkey,
                validator_hotkey=self.validator_hotkey,
                time_started=self.time_started,
                time_took_us=self.time_took_us,
                score_str=self.score_str,
            ),
            validator_signature=self.validator_signature,
            miner_signature=self.miner_signature,
        )



import pytest

from compute_horde.miner_client.organic import (
    OrganicJobDetails,
    OrganicMinerClient,
    run_organic_job,
)
from compute_horde.mv_protocol.miner_requests import (
    V0AcceptJobRequest,
    V0ExecutorReadyRequest,
    V0JobFinishedRequest,
)
from compute_horde.mv_protocol.validator_requests import (
    BaseValidatorRequest,
    V0AuthenticateRequest,
    V0InitialJobRequest,
    V0JobFinishedReceiptRequest,
    V0JobRequest,
    V0JobStartedReceiptRequest,
)
from compute_horde.transport import StubTransport

JOB_UUID = "b4793a02-33a2-4a49-b4e2-4a7b903847e7"


class MinerStubTransport(StubTransport):
    def __init__(self, name: str, messages: list[str], *args, **kwargs):
        super().__init__(name, messages, *args, **kwargs)
        self.sent_models = []

    async def send(self, message):
        await super().send(message)
        self.sent_models.append(BaseValidatorRequest.parse(message))


@pytest.mark.asyncio
async def test_run_organic_job__success(keypair):
    mock_transport = MinerStubTransport(
        "mock",
        [
            V0AcceptJobRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0ExecutorReadyRequest(job_uuid=JOB_UUID).model_dump_json(),
            V0JobFinishedRequest(
                job_uuid=JOB_UUID,
                docker_process_stdout="stdout",
                docker_process_stderr="stderr",
            ).model_dump_json(),
        ],
    )
    client = OrganicMinerClient(
        miner_hotkey="mock",
        miner_address="0.0.0.0",
        miner_port=1234,
        job_uuid=JOB_UUID,
        my_keypair=keypair,
        transport=mock_transport,
    )
    job_details = OrganicJobDetails(job_uuid=JOB_UUID, docker_image="mock")
    stdout, stderr = await run_organic_job(client, job_details, wait_timeout=2)

    assert stdout == "stdout"
    assert stderr == "stderr"

    sent_models_types = [type(model) for model in mock_transport.sent_models]
    assert sent_models_types == [
        V0AuthenticateRequest,
        V0InitialJobRequest,
        V0JobStartedReceiptRequest,
        V0JobRequest,
        V0JobFinishedReceiptRequest,
    ]


# TODO:
#   - unhappy path
#       - connection error
#       - initial_response timeout
#       - send error event callback
#       - declined
#       - executor failed
#       - full_response timeout
#       - job failed



import asyncio

import pytest
import pytest_asyncio
import websockets

from compute_horde.transport import WSTransport


class WSTestServer:
    host = "localhost"
    port = 8765

    def __init__(self):
        self.ws_server = None
        self.received: asyncio.Queue[str] = asyncio.Queue()

    async def srv(self, ws, path):
        async for message in ws:
            await self.received.put(message)

    async def start(self):
        self.ws_server = await websockets.serve(self.srv, self.host, self.port)

    async def stop(self):
        self.ws_server.close()
        await self.ws_server.wait_closed()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()

    @property
    def connection(self) -> websockets.WebSocketServerProtocol | None:
        assert len(self.ws_server.websockets) <= 1

        try:
            return next(iter(self.ws_server.websockets))
        except StopIteration:
            return None

    @property
    def is_connected(self) -> bool:
        return self.connection is not None and self.connection.open


@pytest_asyncio.fixture
async def server():
    async with WSTestServer() as _server:
        yield _server


@pytest.fixture
def ws_transport():
    return WSTransport(
        "test",
        f"ws://{WSTestServer.host}:{WSTestServer.port}",
        base_retry_delay=0.1,
        retry_jitter=0.1,
    )


@pytest.mark.asyncio
async def test_connects(server: WSTestServer, ws_transport: WSTransport):
    assert not server.is_connected

    await ws_transport.start()

    assert server.is_connected


@pytest.mark.asyncio
async def test_send(server: WSTestServer, ws_transport: WSTransport):
    await ws_transport.start()

    await ws_transport.send("foo")
    await ws_transport.send("bar")

    assert await asyncio.wait_for(server.received.get(), 0.2) == "foo"
    assert await asyncio.wait_for(server.received.get(), 0.2) == "bar"


@pytest.mark.asyncio
async def test_send_reconnect(server: WSTestServer, ws_transport: WSTransport):
    await ws_transport.start()
    await server.stop()

    send_task = asyncio.create_task(ws_transport.send("foo"))
    await asyncio.sleep(0.1)
    await server.start()
    await send_task

    assert await asyncio.wait_for(server.received.get(), 0.2) == "foo"


@pytest.mark.asyncio
async def test_receive(server: WSTestServer, ws_transport: WSTransport):
    await ws_transport.start()

    await server.connection.send("foo")
    assert await ws_transport.receive() == "foo"

    await server.connection.send("bar")
    assert await ws_transport.receive() == "bar"


@pytest.mark.asyncio
async def test_receive_reconnect(server: WSTestServer, ws_transport: WSTransport):
    await ws_transport.start()

    await server.stop()

    receive_task = asyncio.create_task(ws_transport.receive())
    await server.start()
    await asyncio.sleep(0.5)
    await server.connection.send("foo")

    assert await receive_task == "foo"



import asyncio
import logging
import random

import websockets

from .base import AbstractTransport, TransportConnectionError

logger = logging.getLogger(__name__)


class WSTransport(AbstractTransport):
    """
    A WebSocket transport layer for exchanging messages between parties.

    An instance needs to be started before sending or receiving messages and stopped after use. That's where the actual connection is established and terminated. Not started instances will raise a runtime error when trying to send or receive messages.

    ```python
    transport = WSTransport("ws://example.com")
    await transport.start()

    await transport.send("Hello, world!")

    message = await transport.receive()

    await transport.stop()
    ```

    Continuously receiving messages can be done using an async for loop.

    ```python
    transport = WSTransport("ws://example.com")
    await transport.start()

    async for message in transport:
        print(message)
    ```
    await transport.stop()
    """

    def __init__(
        self,
        name: str,
        url: str,
        *,
        max_retries: int = 5,
        base_retry_delay: int = 1,
        retry_jitter: float = 1,
    ):
        super().__init__(name)
        self.url = url
        self.base_retry_delay = base_retry_delay
        self.retry_jitter = retry_jitter
        self.max_retries = max_retries
        self.connect_lock = asyncio.Lock()
        self._ws: websockets.WebSocketClientProtocol | None = None

    @property
    def ws(self) -> websockets.WebSocketClientProtocol:
        if self._ws is None:
            raise RuntimeError("WebSocket connection has not been established")
        return self._ws

    def _get_retry_delay(self, attempt: int):
        return self.base_retry_delay * 2**attempt + random.uniform(0, self.retry_jitter)

    async def start(self) -> None:
        async with self.connect_lock:
            await self.connect()

    async def stop(self) -> None:
        async with self.connect_lock:
            if self._ws and self._ws.open:
                await self._ws.close()

    async def connect(self):
        if self._ws and self._ws.open:
            return

        loop = asyncio.get_running_loop()
        start_time = loop.time()
        attempt = 0

        while self.max_retries == 0 or attempt < self.max_retries:
            try:
                self._ws = await websockets.connect(
                    self.url, max_size=50 * (2**20), ping_timeout=120
                )  # 50MB - ping timeout 2min in case of blocking job sends
                logger.info(f"Connected to {self.name} after {attempt} attempts")
                return
            except (websockets.WebSocketException, OSError):
                attempt += 1
                delay = self._get_retry_delay(attempt)
                logger.info(f"Retrying connection to {self.name} in {delay:0.2f}")
                await asyncio.sleep(delay)

        time_took = loop.time() - start_time
        raise TransportConnectionError(
            f"Could not connect to {self.name} after {attempt} attempts"
            f" in {time_took:0.2f} seconds"
        )

    async def send(self, data: str | bytes) -> None:
        while True:
            try:
                await self.ws.send(data)
                await asyncio.sleep(0)
                # Summary: https://github.com/python-websockets/websockets/issues/867
                # Longer discussion: https://github.com/python-websockets/websockets/issues/865
                # logger.debug(f"Sent message to {self.name}: {data}")
                return
            except (websockets.WebSocketException, OSError):
                logger.info(f"Could not send msg to {self.name}. Reconnecting...")
                await self.connect()

    async def receive(self) -> str | bytes:
        while True:
            try:
                msg = await self.ws.recv()
                # logger.debug(f"Received message from {self.name}: {msg}")
                return msg
            except (websockets.WebSocketException, OSError):
                logger.info(f"Could not receive msg from {self.name}. Reconnecting...")
                await self.connect()



import uuid

import bittensor
import pytest

from compute_horde.base_requests import BaseRequest
from compute_horde.miner_client.organic import OrganicMinerClient
from compute_horde.mv_protocol.miner_requests import (
    V0AcceptJobRequest,
    V0DeclineJobRequest,
    V0ExecutorFailedRequest,
    V0ExecutorReadyRequest,
    V0JobFailedRequest,
    V0JobFinishedRequest,
)
from compute_horde.transport import StubTransport

JOB_UUID = "b4793a02-33a2-4a49-b4e2-4a7b903847e7"


def get_miner_client(
    keypair: bittensor.Keypair,
    messages: list[BaseRequest] | None = None,
) -> OrganicMinerClient:
    transport = StubTransport("stub", messages or [])
    client = OrganicMinerClient(
        miner_hotkey="mock",
        miner_address="0.0.0.0",
        miner_port=1234,
        job_uuid=JOB_UUID,
        my_keypair=keypair,
        transport=transport,
    )
    return client


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "initial_msg,executor_msg,final_msg",
    [
        (V0DeclineJobRequest(job_uuid=JOB_UUID), None, None),
        (V0AcceptJobRequest(job_uuid=JOB_UUID), V0ExecutorFailedRequest(job_uuid=JOB_UUID), None),
        (
            V0AcceptJobRequest(job_uuid=JOB_UUID),
            V0ExecutorReadyRequest(job_uuid=JOB_UUID),
            V0JobFailedRequest(
                job_uuid=JOB_UUID,
                docker_process_exit_status=1,
                docker_process_stdout="stdout",
                docker_process_stderr="stderr",
            ),
        ),
        (
            V0AcceptJobRequest(job_uuid=JOB_UUID),
            V0ExecutorReadyRequest(job_uuid=JOB_UUID),
            V0JobFinishedRequest(
                job_uuid=JOB_UUID,
                docker_process_stdout="stdout",
                docker_process_stderr="stderr",
            ),
        ),
    ],
)
async def test_organic_miner_client__futures__properly_set(
    initial_msg, executor_msg, final_msg, keypair
):
    miner_client = get_miner_client(keypair)

    assert not miner_client.miner_accepting_or_declining_future.done()
    assert miner_client.miner_accepting_or_declining_timestamp == 0

    await miner_client.handle_message(initial_msg)

    assert miner_client.miner_accepting_or_declining_future.done()
    assert miner_client.miner_accepting_or_declining_timestamp != 0
    assert await miner_client.miner_accepting_or_declining_future == initial_msg

    # should set only once
    set_time = miner_client.miner_accepting_or_declining_timestamp
    await miner_client.handle_message(initial_msg)
    assert miner_client.miner_accepting_or_declining_timestamp == set_time

    if executor_msg is None:
        return

    assert not miner_client.executor_ready_or_failed_future.done()
    assert miner_client.executor_ready_or_failed_timestamp == 0

    await miner_client.handle_message(executor_msg)

    assert miner_client.executor_ready_or_failed_future.done()
    assert miner_client.executor_ready_or_failed_future != 0
    assert await miner_client.executor_ready_or_failed_future == executor_msg

    # should set only once
    set_time = miner_client.executor_ready_or_failed_timestamp
    await miner_client.handle_message(executor_msg)
    assert miner_client.executor_ready_or_failed_timestamp == set_time

    if final_msg is None:
        return

    assert not miner_client.miner_finished_or_failed_future.done()
    assert miner_client.miner_finished_or_failed_timestamp == 0

    await miner_client.handle_message(final_msg)

    assert miner_client.miner_finished_or_failed_future.done()
    assert miner_client.miner_finished_or_failed_timestamp != 0
    assert await miner_client.miner_finished_or_failed_future == final_msg

    # should set only once
    set_time = miner_client.miner_finished_or_failed_timestamp
    await miner_client.handle_message(final_msg)
    assert miner_client.miner_finished_or_failed_timestamp == set_time


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "initial_msg",
    [
        V0AcceptJobRequest(job_uuid=str(uuid.uuid4())),
        V0DeclineJobRequest(job_uuid=str(uuid.uuid4())),
    ],
)
async def test_organic_miner_client__skip_different_job__initial_future(initial_msg, keypair):
    miner_client = get_miner_client(keypair)

    assert not miner_client.miner_accepting_or_declining_future.done()
    assert miner_client.miner_accepting_or_declining_timestamp == 0

    await miner_client.handle_message(initial_msg)

    assert not miner_client.miner_accepting_or_declining_future.done()
    assert miner_client.miner_accepting_or_declining_timestamp == 0


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "final_msg",
    [
        V0JobFailedRequest(
            job_uuid=str(uuid.uuid4()),
            docker_process_exit_status=1,
            docker_process_stdout="stdout",
            docker_process_stderr="stderr",
        ),
        V0JobFinishedRequest(
            job_uuid=str(uuid.uuid4()),
            docker_process_stdout="stdout",
            docker_process_stderr="stderr",
        ),
    ],
)
async def test_organic_miner_client__skip_different_job__final_future(final_msg, keypair):
    miner_client = get_miner_client(keypair)

    assert not miner_client.miner_finished_or_failed_future.done()
    assert miner_client.miner_finished_or_failed_timestamp == 0

    await miner_client.handle_message(final_msg)

    assert not miner_client.miner_finished_or_failed_future.done()
    assert miner_client.miner_finished_or_failed_timestamp == 0



import datetime
import logging
from typing import Any, Protocol

import requests
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class ParamItem(BaseModel):
    value: Any
    effective_from: datetime.datetime | None = None  # None = from the beginning of time (-inf)
    reason: str | None = None


class Param(BaseModel):
    description: str
    items: list[ParamItem]


class SupportsSetAttr(Protocol):
    def __setattr__(self, key: str, value: Any) -> None: ...


def sync_dynamic_config(config_url: str, namespace: SupportsSetAttr) -> None:
    """
    Fetches dynamic config from a URL and sets them to namespace.
    Only the keys with `DYNAMIC_` prefix are considered.

    :param config_url: URL to a JSON object of dynamic config
    :param namespace: An object where the config values will be set as attributes
    """
    response = requests.get(config_url)
    response.raise_for_status()
    for param_key, raw_param in response.json().items():
        if not param_key.startswith("DYNAMIC_"):
            continue

        param = Param.model_validate(raw_param)
        now = datetime.datetime.now(datetime.UTC)
        for param_item in param.items:
            if param_item.effective_from is not None and param_item.effective_from > now:
                logger.warning(f"Ignoring dynamic config {param_key}={param_item.value}")
                continue

            try:
                setattr(namespace, param_key, param_item.value)
            except AttributeError:
                logger.warning(f"Failed to set dynamic config {param_key}={param_item.value}")
            else:
                msg = f"Set dynamic config {param_key}={param_item.value}"
                if param_item.reason:
                    msg += f", reason={param_item.reason}"
                logger.info(msg)



from compute_horde.receipts.schemas import Receipt, ReceiptType
from compute_horde.receipts.transfer import ReceiptFetchError, get_miner_receipts

default_app_config = "compute_horde.receipts.apps.ComputeHordeReceiptsConfig"

# Reexported for compatibility. These were moved to submodules.
__all__ = [
    "Receipt",
    "ReceiptType",
    "get_miner_receipts",
    "ReceiptFetchError",
]



import enum
import logging

import bittensor
import pydantic

from compute_horde.mv_protocol.validator_requests import (
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
)

logger = logging.getLogger(__name__)


class ReceiptType(enum.Enum):
    JobStartedReceipt = "JobStartedReceipt"
    JobFinishedReceipt = "JobFinishedReceipt"


class Receipt(pydantic.BaseModel):
    payload: JobStartedReceiptPayload | JobFinishedReceiptPayload
    validator_signature: str
    miner_signature: str

    def verify_miner_signature(self):
        miner_keypair = bittensor.Keypair(ss58_address=self.payload.miner_hotkey)
        return miner_keypair.verify(self.payload.blob_for_signing(), self.miner_signature)

    def verify_validator_signature(self):
        validator_keypair = bittensor.Keypair(ss58_address=self.payload.validator_hotkey)
        return validator_keypair.verify(self.payload.blob_for_signing(), self.validator_signature)



import datetime
from typing import TYPE_CHECKING, Any

import bittensor
import pydantic
from substrateinterface.exceptions import SubstrateRequestException

if TYPE_CHECKING:
    from bittensor.chain_data import NeuronInfo

BAC_VALIDATOR_SS58_ADDRESS = "5HBVrFGy6oYhhh71m9fFGYD7zbKyAeHnWN8i8s9fJTBMCtEE"
MIN_STAKE = 1000
VALIDATORS_LIMIT = 24


class MachineSpecs(pydantic.BaseModel):
    specs: dict[Any, Any]

    def __str__(self) -> str:
        return str(self.specs)


class ValidatorListError(Exception):
    def __init__(self, reason: Exception):
        self.reason = reason


def get_validators(netuid=12, network="finney", block: int | None = None) -> list["NeuronInfo"]:
    """
    Validators are top 24 neurons in terms of stake, only taking into account those that have at least 1000
    and forcibly including BAC_VALIDATOR_SS58_ADDRESS.
    The result is sorted.
    """
    try:
        subtensor = bittensor.subtensor(network=network)
    except Exception as ex:
        raise ValidatorListError(ex) from ex

    try:
        metagraph = subtensor.metagraph(netuid, block=block)
    except SubstrateRequestException as ex:
        raise ValidatorListError(ex) from ex

    neurons = [
        n
        for n in metagraph.neurons
        if (n.hotkey == BAC_VALIDATOR_SS58_ADDRESS or n.stake.tao >= MIN_STAKE)
    ]
    neurons = sorted(
        neurons, key=lambda n: (n.hotkey == BAC_VALIDATOR_SS58_ADDRESS, n.stake), reverse=True
    )
    return neurons[:VALIDATORS_LIMIT]


def _json_dumps_default(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    raise TypeError


class Timer:
    def __init__(self, timeout=None):
        self.start_time = datetime.datetime.now()
        self.timeout = timeout

    def passed_time(self):
        return (datetime.datetime.now() - self.start_time).total_seconds()

    def time_left(self):
        if self.timeout is None:
            raise ValueError("timeout was not specified")
        return self.timeout - self.passed_time()



import asyncio
import contextlib
import datetime
import enum
import logging
import time
from dataclasses import dataclass, field
from functools import cached_property

import bittensor

from compute_horde.base.docker import DockerRunOptionsPreset
from compute_horde.base.output_upload import OutputUpload
from compute_horde.base.volume import Volume
from compute_horde.base_requests import BaseRequest
from compute_horde.executor_class import ExecutorClass
from compute_horde.miner_client.base import (
    AbstractMinerClient,
    ErrorCallback,
    UnsupportedMessageReceived,
)
from compute_horde.mv_protocol import miner_requests, validator_requests
from compute_horde.mv_protocol.miner_requests import (
    BaseMinerRequest,
    UnauthorizedError,
    V0AcceptJobRequest,
    V0DeclineJobRequest,
    V0ExecutorFailedRequest,
    V0ExecutorManifestRequest,
    V0ExecutorReadyRequest,
    V0JobFailedRequest,
    V0JobFinishedRequest,
    V0MachineSpecsRequest,
)
from compute_horde.mv_protocol.validator_requests import (
    AuthenticationPayload,
    JobFinishedReceiptPayload,
    JobStartedReceiptPayload,
    V0AuthenticateRequest,
    V0InitialJobRequest,
    V0JobFinishedReceiptRequest,
    V0JobRequest,
    V0JobStartedReceiptRequest,
)
from compute_horde.transport import AbstractTransport, TransportConnectionError, WSTransport
from compute_horde.utils import MachineSpecs, Timer

logger = logging.getLogger(__name__)


class OrganicMinerClient(AbstractMinerClient):
    """
    Miner client to run organic job on a miner.
    This client is used by validators to connect to prod miners.
    Others can use this client to connect to their locally running miners to run jobs.

    The usual usage of this miner client is as follows::

        client = OrganicMinerClient(...)
        async with client:
            await client.send_model(V0InitialJobRequest(...))
            msg = await client.miner_ready_or_declining_future
            # handle msg

            await client.send_model(V0JobRequest(...))
            msg = client.miner_finished_or_failed_future
            # handle msg

    Note that the waiting on the response futures should properly be handled with timeouts
    (with ``asyncio.timeout()``, ``asyncio.wait_for()`` etc.).
    """

    def __init__(
        self,
        miner_hotkey: str,
        miner_address: str,
        miner_port: int,
        job_uuid: str,
        my_keypair: bittensor.Keypair,
        transport: AbstractTransport | None = None,
    ) -> None:
        self.job_uuid = job_uuid

        self.miner_hotkey = miner_hotkey
        self.miner_address = miner_address
        self.miner_port = miner_port
        self.my_keypair = my_keypair

        loop = asyncio.get_running_loop()
        self.miner_manifest = loop.create_future()
        self.online_executor_count = 0

        # for waiting on miner responses
        self.miner_accepting_or_declining_future: asyncio.Future[
            V0AcceptJobRequest | V0DeclineJobRequest
        ] = loop.create_future()
        self.miner_accepting_or_declining_timestamp: int = 0

        self.executor_ready_or_failed_future: asyncio.Future[
            V0ExecutorReadyRequest | V0ExecutorFailedRequest
        ] = loop.create_future()
        self.executor_ready_or_failed_timestamp: int = 0

        self.miner_finished_or_failed_future: asyncio.Future[
            V0JobFailedRequest | V0JobFinishedRequest
        ] = loop.create_future()
        self.miner_finished_or_failed_timestamp: int = 0

        self.miner_machine_specs: MachineSpecs | None = None

        name = f"{miner_hotkey}({miner_address}:{miner_port})"
        transport = transport or WSTransport(name, self.miner_url())
        super().__init__(name, transport)

    @cached_property
    def my_hotkey(self) -> str:
        return str(self.my_keypair.ss58_address)

    def miner_url(self) -> str:
        return (
            f"ws://{self.miner_address}:{self.miner_port}/v0.1/validator_interface/{self.my_hotkey}"
        )

    def accepted_request_type(self) -> type[BaseRequest]:
        return BaseMinerRequest

    def incoming_generic_error_class(self) -> type[BaseRequest]:
        return miner_requests.GenericError

    def outgoing_generic_error_class(self) -> type[BaseRequest]:
        return validator_requests.GenericError

    def build_outgoing_generic_error(self, msg: str):
        return validator_requests.GenericError(details=msg)

    async def notify_generic_error(self, msg: BaseRequest) -> None:
        """This method is called when miner sends a generic error message"""

    async def notify_unauthorized_error(self, msg: UnauthorizedError) -> None:
        """This method is called when miner refuses the authentication message"""

    async def notify_receipt_failure(self, comment: str) -> None:
        """This method is called when sending receipts to miner fails"""

    async def notify_send_failure(self, msg: str) -> None:
        """This method is called when sending messages to miner fails"""

    async def handle_manifest_request(self, msg: V0ExecutorManifestRequest) -> None:
        try:
            self.miner_manifest.set_result(msg.manifest)
        except asyncio.InvalidStateError:
            logger.warning(f"Received manifest from {msg} but future was already set")

    async def handle_machine_specs_request(self, msg: V0MachineSpecsRequest) -> None:
        self.miner_machine_specs = msg.specs

    async def handle_message(self, msg: BaseRequest) -> None:
        if isinstance(msg, self.incoming_generic_error_class()):
            logger.warning(
                f"Received error message from miner {self.miner_name}: {msg.model_dump_json()}"
            )
            await self.notify_generic_error(msg)
            return
        elif isinstance(msg, UnauthorizedError):
            logger.error(f"Unauthorized in {self.miner_name}: {msg.code}, details: {msg.details}")
            await self.notify_unauthorized_error(msg)
            return
        elif isinstance(msg, V0ExecutorManifestRequest):
            await self.handle_manifest_request(msg)
            return

        if (received_job_uuid := getattr(msg, "job_uuid", self.job_uuid)) != self.job_uuid:
            logger.warning(
                f"Received msg from {self.miner_name} for a different job (expected job {self.job_uuid}, got job {received_job_uuid}): {msg}"
            )
            return

        if isinstance(msg, V0AcceptJobRequest | V0DeclineJobRequest):
            try:
                self.miner_accepting_or_declining_future.set_result(msg)
                self.miner_accepting_or_declining_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0ExecutorReadyRequest | V0ExecutorFailedRequest):
            try:
                self.executor_ready_or_failed_future.set_result(msg)
                self.executor_ready_or_failed_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0JobFailedRequest | V0JobFinishedRequest):
            try:
                self.miner_finished_or_failed_future.set_result(msg)
                self.miner_finished_or_failed_timestamp = int(time.time())
            except asyncio.InvalidStateError:
                logger.warning(f"Received {msg} from {self.miner_name} but future was already set")
        elif isinstance(msg, V0MachineSpecsRequest):
            await self.handle_machine_specs_request(msg)
        else:
            raise UnsupportedMessageReceived(msg)

    def generate_authentication_message(self) -> V0AuthenticateRequest:
        payload = AuthenticationPayload(
            validator_hotkey=self.my_hotkey,
            miner_hotkey=self.miner_hotkey,
            timestamp=int(time.time()),
        )
        return V0AuthenticateRequest(
            payload=payload, signature=f"0x{self.my_keypair.sign(payload.blob_for_signing()).hex()}"
        )

    def generate_job_started_receipt_message(
        self,
        executor_class: ExecutorClass,
        accepted_timestamp: float,
        max_timeout: int,
    ) -> V0JobStartedReceiptRequest:
        time_accepted = datetime.datetime.fromtimestamp(accepted_timestamp, datetime.UTC)
        receipt_payload = JobStartedReceiptPayload(
            job_uuid=self.job_uuid,
            miner_hotkey=self.miner_hotkey,
            validator_hotkey=self.my_hotkey,
            executor_class=executor_class,
            time_accepted=time_accepted,
            max_timeout=max_timeout,
        )
        return V0JobStartedReceiptRequest(
            payload=receipt_payload,
            signature=f"0x{self.my_keypair.sign(receipt_payload.blob_for_signing()).hex()}",
        )

    async def send_job_started_receipt_message(
        self,
        executor_class: ExecutorClass,
        accepted_timestamp: float,
        max_timeout: int,
    ) -> None:
        try:
            receipt_message = self.generate_job_started_receipt_message(
                executor_class,
                accepted_timestamp,
                max_timeout,
            )
            await self.send_model(receipt_message)
            logger.debug(f"Sent job started receipt for {self.job_uuid}")
        except Exception as e:
            comment = f"Failed to send job started receipt to miner {self.miner_name} for job {self.job_uuid}: {e}"
            logger.warning(comment)
            await self.notify_receipt_failure(comment)

    def generate_job_finished_receipt_message(
        self,
        started_timestamp: float,
        time_took_seconds: float,
        score: float,
    ) -> V0JobFinishedReceiptRequest:
        time_started = datetime.datetime.fromtimestamp(started_timestamp, datetime.UTC)
        receipt_payload = JobFinishedReceiptPayload(
            job_uuid=self.job_uuid,
            miner_hotkey=self.miner_hotkey,
            validator_hotkey=self.my_hotkey,
            time_started=time_started,
            time_took_us=int(time_took_seconds * 1_000_000),
            score_str=f"{score:.6f}",
        )
        return V0JobFinishedReceiptRequest(
            payload=receipt_payload,
            signature=f"0x{self.my_keypair.sign(receipt_payload.blob_for_signing()).hex()}",
        )

    async def send_job_finished_receipt_message(
        self,
        started_timestamp: float,
        time_took_seconds: float,
        score: float,
    ) -> None:
        try:
            receipt_message = self.generate_job_finished_receipt_message(
                started_timestamp, time_took_seconds, score
            )
            await self.send_model(receipt_message)
            logger.debug(f"Sent job finished receipt for {self.job_uuid}")
        except Exception as e:
            comment = f"Failed to send job finished receipt to miner {self.miner_name} for job {self.job_uuid}: {e}"
            logger.warning(comment)
            await self.notify_receipt_failure(comment)

    async def send_model(
        self, model: BaseRequest, error_event_callback: ErrorCallback | None = None
    ) -> None:
        if error_event_callback is None:
            error_event_callback = self.notify_send_failure

        await super().send_model(model, error_event_callback)

    async def connect(self) -> None:
        await super().connect()
        await self.transport.send(self.generate_authentication_message().model_dump_json())


class FailureReason(enum.Enum):
    MINER_CONNECTION_FAILED = enum.auto()
    INITIAL_RESPONSE_TIMED_OUT = enum.auto()
    EXECUTOR_READINESS_RESPONSE_TIMED_OUT = enum.auto()
    FINAL_RESPONSE_TIMED_OUT = enum.auto()
    JOB_DECLINED = enum.auto()
    EXECUTOR_FAILED = enum.auto()
    JOB_FAILED = enum.auto()


class OrganicJobError(Exception):
    def __init__(self, reason: FailureReason, received: BaseRequest | None = None):
        self.reason = reason
        self.received = received

    def __str__(self):
        s = f"Organic job failed, {self.reason=}"
        if self.received:
            s += f", {self.received=}"
        return s

    def __repr__(self):
        return f"{type(self).__name__}: {str(self)}"


@dataclass
class OrganicJobDetails:
    job_uuid: str
    executor_class: ExecutorClass = ExecutorClass.spin_up_4min__gpu_24gb
    docker_image: str | None = None
    raw_script: str | None = None
    docker_run_options_preset: DockerRunOptionsPreset = "nvidia_all"
    docker_run_cmd: list[str] = field(default_factory=list)
    total_job_timeout: int = 300
    volume: Volume | None = None
    output: OutputUpload | None = None

    def __post_init__(self):
        if (self.docker_image, self.raw_script) == (None, None):
            raise ValueError("At least of of `docker_image` or `raw_script` must be not None")


async def run_organic_job(
    client: OrganicMinerClient,
    job_details: OrganicJobDetails,
    wait_timeout: int = 300,
):
    """
    Run an organic job. This is a simpler way to use OrganicMinerClient.

    :param client: the organic miner client
    :param job_details: details specific to the job that needs to be run
    :param wait_timeout: maximum timeout for waiting for miner responses
    :return: standard out and standard error of the job container
    """
    assert client.job_uuid == job_details.job_uuid

    async with contextlib.AsyncExitStack() as exit_stack:
        try:
            await exit_stack.enter_async_context(client)
        except TransportConnectionError as exc:
            raise OrganicJobError(FailureReason.MINER_CONNECTION_FAILED) from exc

        job_timer = Timer(timeout=job_details.total_job_timeout)

        await client.send_model(
            V0InitialJobRequest(
                job_uuid=job_details.job_uuid,
                executor_class=job_details.executor_class,
                base_docker_image_name=job_details.docker_image,
                timeout_seconds=job_details.total_job_timeout,
                volume_type=job_details.volume.volume_type if job_details.volume else None,
            ),
        )

        try:
            initial_response = await asyncio.wait_for(
                client.miner_accepting_or_declining_future,
                timeout=min(job_timer.time_left(), wait_timeout),
            )
        except TimeoutError as exc:
            raise OrganicJobError(FailureReason.INITIAL_RESPONSE_TIMED_OUT) from exc
        if isinstance(initial_response, V0DeclineJobRequest):
            raise OrganicJobError(FailureReason.JOB_DECLINED, initial_response)

        try:
            executor_readiness_response = await asyncio.wait_for(
                client.executor_ready_or_failed_future,
                timeout=min(job_timer.time_left(), wait_timeout),
            )
        except TimeoutError as exc:
            raise OrganicJobError(FailureReason.EXECUTOR_READINESS_RESPONSE_TIMED_OUT) from exc
        if isinstance(executor_readiness_response, V0ExecutorFailedRequest):
            raise OrganicJobError(FailureReason.EXECUTOR_FAILED, initial_response)

        await client.send_job_started_receipt_message(
            executor_class=job_details.executor_class,
            accepted_timestamp=time.time(),
            max_timeout=int(job_timer.time_left()),
        )

        await client.send_model(
            V0JobRequest(
                job_uuid=job_details.job_uuid,
                executor_class=job_details.executor_class,
                docker_image_name=job_details.docker_image,
                raw_script=job_details.raw_script,
                docker_run_options_preset=job_details.docker_run_options_preset,
                docker_run_cmd=job_details.docker_run_cmd,
                volume=job_details.volume,
                output_upload=job_details.output,
            )
        )

        try:
            final_response = await asyncio.wait_for(
                client.miner_finished_or_failed_future,
                timeout=job_timer.time_left(),
            )
            if isinstance(final_response, V0JobFailedRequest):
                raise OrganicJobError(FailureReason.JOB_FAILED, final_response)
            return final_response.docker_process_stdout, final_response.docker_process_stderr
        except TimeoutError as exc:
            raise OrganicJobError(FailureReason.FINAL_RESPONSE_TIMED_OUT) from exc
        finally:
            await client.send_job_finished_receipt_message(
                started_timestamp=job_timer.start_time.timestamp(),
                time_took_seconds=job_timer.passed_time(),
                score=0,  # no score for organic jobs (at least right now)
            )



import asyncio

from .base import AbstractTransport


class StubTransport(AbstractTransport):
    """
    Stub transport for testing purposes.
    Receives predefined list of messages and saves sent messages in the internal state
    """

    def __init__(self, name: str, messages: list[str], *args, **kwargs):
        self.sent: list[str] = []
        self.messages = iter(messages)
        self.sent_messages: list[str] = []

    async def start(self): ...

    async def stop(self): ...

    async def send(self, message):
        self.sent_messages.append(message)

    async def receive(self):
        try:
            return next(self.messages)
        except StopIteration:
            await asyncio.Future()
